from __future__ import unicode_literals

from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class FormDesignerConfig(AppConfig):
    name = 'form_designer'
    verbose_name = _("Form Designer")
